library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

entity mousedriver is
    Generic( timeout: std_logic_vector(19 downto 0) := CONV_STD_LOGIC_VECTOR(100000,20));
    Port ( clock : in std_logic;
           reset : in std_logic;
           mouse_data : inout std_logic;
           mouse_clk : inout std_logic;
           left_button : out std_logic;
           right_button : out std_logic;
           mouse_cursor_row : out std_logic_vector(9 downto 0);
           mouse_cursor_column : out std_logic_vector(9 downto 0));
end mousedriver;

architecture Behavioral of mousedriver is
type state_type is (POWERON, INIT, SENDBYTE, GETBYTE);
type byte_id is (GETACKBYTE, GETBYTE1, GETBYTE2, GETBYTE3);

signal clk_filter: std_logic_vector(3 downto 0);
signal outbyte: std_logic_vector(10 downto 0); -- with stopbit and parity
signal inbyte: std_logic_vector(7 downto 0);
signal byte1: std_logic_vector(7 downto 0);
signal byte2: std_logic_vector(7 downto 0);
signal bitcnt: std_logic_vector(3 downto 0);
signal clk_filtered: std_logic;
signal parity: std_logic;

signal state: state_type := POWERON;
signal whichbyte: byte_id;
signal initcounter: std_logic_vector(10 downto 0);

signal mouse_data_dir: std_logic; --1 - in, 0 - out
signal mouse_clk_dir: std_logic;
signal mouse_data_i: std_logic;
signal mouse_clk_i: std_logic;
signal mouse_cursor_row_int : std_logic_vector(9 downto 0);
signal mouse_cursor_column_int : std_logic_vector(9 downto 0);

signal watchdog: std_logic_vector(19 downto 0);

begin

  mouse_clk <= mouse_clk_i when (mouse_clk_dir='0') else 'Z';
  mouse_data <= mouse_data_i when (mouse_data_dir='0') else 'Z';

  mouse_cursor_row <= mouse_cursor_row_int;
  mouse_cursor_column <= mouse_cursor_column_int;

--  mouse_cursor_column <= CONV_STD_LOGIC_VECTOR(0,10) when state = POWERON else
--                         CONV_STD_LOGIC_VECTOR(1,10) when state = INIT else                         
--                         CONV_STD_LOGIC_VECTOR(2,10) when state = SENDBYTE else
--                         CONV_STD_LOGIC_VECTOR(3,10);                          

process (clock)
variable clk_filter_i: std_logic_vector(3 downto 0);
variable mouse_cursor_row_i : std_logic_vector(9 downto 0);
variable mouse_cursor_column_i : std_logic_vector(9 downto 0);
variable watchdog_expired : std_logic;
begin
if clock'event and clock='1' then
   
   watchdog <= watchdog - 1;
   watchdog_expired := '0';
   if watchdog = CONV_STD_LOGIC_VECTOR(0,20) then  
     if state = SENDBYTE or 
        ( state = GETBYTE and (whichbyte = GETACKBYTE or whichbyte = GETBYTE2 or whichbyte = GETBYTE3
                                or (whichbyte = GETBYTE1 and bitcnt /= CONV_STD_LOGIC_VECTOR(10,4)) ) )
           then
       watchdog_expired := '1';
     end if;
   end if;
      
   if (watchdog_expired = '1') or (reset = '1') or (state = POWERON) then
    state <= INIT;
    initcounter <= CONV_STD_LOGIC_VECTOR(1500,11);
    mouse_cursor_row_int <= CONV_STD_LOGIC_VECTOR(240,10);
    mouse_cursor_column_int <= CONV_STD_LOGIC_VECTOR(320,10);
   elsif state = INIT then
    mouse_clk_dir <= '0';
    mouse_data_dir <= '0';
    mouse_clk_i <= '0'; -- stop transmission
    mouse_data_i <= '0'; -- start bit
    if initcounter = CONV_STD_LOGIC_VECTOR(1,11) then
      mouse_clk_i <= '1'; -- move the line up
    end if;
    if initcounter = CONV_STD_LOGIC_VECTOR(0,11) then
      mouse_clk_i <= '1';
      mouse_clk_dir <= '1'; -- and then tristate - simulate the open drain output
      outbyte <= "11011110100"; -- send 0xf4, stop, parity as msb, two stop bits necessary 
                               -- last clock is for mouse sensing, if we have more bytes to send
      state <= SENDBYTE;
      watchdog <= timeout;
      clk_filtered <= '1';
      clk_filter <= "1111";
      bitcnt <= CONV_STD_LOGIC_VECTOR(10,4); -- 11 bits to send
    end if;
    initcounter <= initcounter - 1;
   else -- not reset, not POWERON and not INIT
    clk_filter_i(3 downto 1) := clk_filter(2 downto 0);
    clk_filter_i(0):=mouse_clk;
    clk_filter <= clk_filter_i;
    if clk_filter_i="1111" then
      clk_filtered <= '1';
    end if;
    if clk_filter_i = "0000" then
      clk_filtered <= '0';   
      if clk_filtered = '1' then
         -- falling edge of clk_filtered
         case state is
           when SENDBYTE =>
                      mouse_data_i <= outbyte(0);
                      outbyte(9 downto 0) <= outbyte(10 downto 1);
                      if bitcnt = "0000" then
                        bitcnt <= CONV_STD_LOGIC_VECTOR(10,4); -- 11 bits to receive
                        state <= GETBYTE;
                        whichbyte <= GETACKBYTE;
                        watchdog <= timeout;
                        mouse_data_dir <= '1';
                      else
                        bitcnt <= bitcnt - 1;
                      end if;
           when GETBYTE =>
                   case bitcnt is
                     when "1010" => -- 10
                       if mouse_data = '0' then -- got start bit
                         if whichbyte=GETBYTE1 then
                           watchdog <= timeout;
                         end if;
                         bitcnt <= bitcnt - 1;
                         parity <= '0';
                       end if;
                     when "0001" =>
                       if (parity xor mouse_data) = '1' then
                         bitcnt <= bitcnt - 1;
                        else 
                          state <= POWERON;
                       end if;
                     when "0000" =>
                       if mouse_data = '1' then -- stop bit
                         bitcnt <= CONV_STD_LOGIC_VECTOR(10,4);
                         case whichbyte is
                          when GETACKBYTE =>
                           if inbyte = X"fa" then -- ack
                             whichbyte <= GETBYTE1;
                           else
                             state <= POWERON;
                           end if;
                          when GETBYTE1 =>
                           whichbyte <= GETBYTE2;
                           byte1 <= inbyte;
                           if inbyte(3) = '0' then
                             state <= POWERON;
                           watchdog <= timeout;
                           end if;
                          when GETBYTE2 =>
                           whichbyte <= GETBYTE3;
                           byte2 <= inbyte;
                           watchdog <= timeout;
                          when GETBYTE3 =>
                           whichbyte <= GETBYTE1;
                           -- now do something with the bytes
                           left_button <= byte1(0);
                           right_button <= byte1(1);
                           if byte2(7) = '1' then
                             mouse_cursor_column_i := mouse_cursor_column_int + byte2(6 downto 0) - 128;
                             if mouse_cursor_column_i>=640 then
                                mouse_cursor_column_i := CONV_STD_LOGIC_VECTOR(0,10);
                             end if;
                           else
                             mouse_cursor_column_i := mouse_cursor_column_int + byte2(6 downto 0);
                             if mouse_cursor_column_i>=640 then
                                mouse_cursor_column_i := CONV_STD_LOGIC_VECTOR(639,10);
                            end if;
                           end if;
            
                           if inbyte(7) = '1' then
                             mouse_cursor_row_i := mouse_cursor_row_int - inbyte(6 downto 0) + 128;
                             if mouse_cursor_row_i>=480 then
                                mouse_cursor_row_i := CONV_STD_LOGIC_VECTOR(479,10);
                             end if;
                           else
                             mouse_cursor_row_i := mouse_cursor_row_int - inbyte(6 downto 0);
                             if mouse_cursor_row_i>=480 then
                                mouse_cursor_row_i := CONV_STD_LOGIC_VECTOR(0,10);
                             end if;
                           end if;
                           mouse_cursor_column_int <= mouse_cursor_column_i;
                           mouse_cursor_row_int <= mouse_cursor_row_i;
                         end case;
                       else -- stop bit missing
                          state <= POWERON;
                       end if;
                     when others =>
                       inbyte(6 downto 0) <= inbyte(7 downto 1);
                       inbyte(7) <= mouse_data;
                       parity <= parity xor mouse_data;
                       bitcnt <= bitcnt - 1;
                   end case;
           when others =>
             -- empty
         end case;
      end if; 
    end if;
   end if;

end if;
end process;

end Behavioral;
  
